==> nono-0.7.2.tar.gz: nono-0.7.2/vm/planevram.cpp <==
//
// nono
// Copyright (C) 2022 nono project
// Licensed under nono-license.txt
//

//
// プレーン VRAM (垂直 VRAM)
// Lunafb と X68k TVRAM の共通部分
//

#include "planevram.h"
#include "mybswap.h"

// コンストラクタ
PlaneVRAMDevice::PlaneVRAMDevice(int width_, int height_)
	: inherited(OBJ_PLANEVRAM)
{
	// ポカ避け。継承クラス側が設定する。
	nplane = -1;

	composite.Create(width_, height_);

	dirty.resize(height_);

	// レンダリング時に使うテーブルを事前に計算。
	InitDeptable();

#if defined(HAVE_AVX2)
	enable_avx2 = gMainApp.enable_avx2;
#endif
}

// デストラクタ
PlaneVRAMDevice::~PlaneVRAMDevice()
{
}

// リセット
void
PlaneVRAMDevice::ResetHard(bool poweron)
{
	// XXX ここ?
	Invalidate();
}

// レンダリング時に使うテーブルを事前に計算。
void
PlaneVRAMDevice::InitDeptable()
{
	deptable.reset(new uint64[256]);

	// %abcdefgh の8ビットを
	// リトルエンディアンホストでは
	// %0000000h'0000000g'0000000f'0000000e'0000000d'0000000c'0000000b'0000000a
	// ビッグエンディアンホストでは
	// %0000000a'0000000b'0000000c'0000000d'0000000e'0000000f'0000000g'0000000h
	// の64ビットに伸張する。
	// (8ビットを各バイトの最下位ビットに対応させる)
	for (int i = 0; i < 256; i++) {
		uint64 res = 0;
		for (int bit = 0; bit < 8; bit++) {
			int n = bit * 8;
#if BYTE_ORDER == LITTLE_ENDIAN
			n = 56 - n;
#endif
			res |= (uint64)((i >> bit) & 1) << n;
		}
		deptable[i] = res;
	}
}

// 全画面の更新フラグを立てる
void
PlaneVRAMDevice::Invalidate()
{
	std::fill(dirty.begin(), dirty.end(), 1);
}

// 全画面の更新フラグを下ろす
// VM スレッドから呼ばれる。
void
PlaneVRAMDevice::ClearDirty()
{
	std::fill(dirty.begin(), dirty.end(), 0);
}

// 更新情報を取得。
// VM スレッドから呼ばれる。
ModifyInfo
PlaneVRAMDevice::GetModify() const
{
	ModifyInfo mod;
	memcpy(&mod.bits[0], &dirty[0], mod.bits.size());
	// TODO: 互換性のために n_dirty 計算。廃止するかも。
	for (int i = 0; i < dirty.size(); i++) {
		if (dirty[i]) {
			mod.n_dirty++;
		}
	}

	return mod;
}

// 画面合成。
// レンダラスレッドから呼ばれる。
bool
PlaneVRAMDevice::Render(BitmapRGBX& dst, const ModifyInfo& modify)
{
	bool updated = false;

	// VRAM に更新があれば composite を更新
	if (modify.IsDirty()) {
		RenderVRAMToComposite(modify);
		updated = true;
	}

	// composite に更新があるか(update)、
	// 無条件に更新するか(invalidate2) なら dst を更新。
	if (updated || modify.invalidate2) {
		RenderCompositeToRGBX(dst, modify);
		updated = true;
	}

	return updated;
}

// VRAM から composite 画面を合成する。
// レンダラスレッドから呼ばれる。
void
PlaneVRAMDevice::RenderVRAMToComposite(const ModifyInfo& modify)
{
	const int width  = composite.GetWidth();
	const int height = composite.GetHeight();
	const int planesz = (width / 8) * height;

	// 1bpp はパレット的にはほぼ 4bpp と同じ扱いになる。
	const int nplane4 = (nplane == 1) ? 4 : nplane;

	// テキスト合成画面を描く
	for (int y = 0; y < height; y++) {
		uint8 mod = modify.bits[y];
		if (mod == 0) {
			continue;
		}

		const uint32 *src = (const uint32 *)&mem[0];
		src += y * width / 32;
		uint64 *dst = (uint64 *)composite.GetRowPtr(y);

		for (int x = 0; x < width / 32; x++) {
			std::array<uint32, 8> data;

			// VRAM はホストエンディアンに関係なく 32bit で読み込んだ
			// 時の MSB が左端ピクセル。
			// ここでは左側から 8ピクセル(8ビット)ずつ 4回処理する
			// のを最下位バイトから順に右シフトで行いたいので、
			// (バイト内のビット順は維持したまま) バイトスワップする。
			//
			//          31     24  23     16  15      8   7      0
			//        +----------+----------+----------+----------+
			// VRAM   |+00 .. +07|+08 .. +15|+16 .. +23|+24 .. +31|
			//        +----------+----------+----------+----------+
			//  ↓
			//        +----------+----------+----------+----------+
			// data[] |+24 .. +31|+16 .. +23|+08 .. +15|+00 .. +07|
			//        +----------+----------+----------+----------+
			if (__predict_false(nplane == 1)) {
				// LUNA 1bpp の場合残りの3プレーンは ff で埋める。
				// 実際にも存在しないプレーンは ff が読み出せて
				// カラーインデックスは $e か $f の二択になるので。
				data[0] = bswap32(src[0]);
				data[1] = 0xffffffff;
				data[2] = 0xffffffff;
				data[3] = 0xffffffff;
			} else {
				int offset = planesz / sizeof(uint32);
				for (int plane = 0; plane < nplane; plane++) {
					data[plane] = bswap32(src[plane * offset]);
				}
			}
			src++;

			// data[] の下位 8bit ずつを取り出して deptable[] で変換。
			// p0..p3 は1バイト x 8個。
			//
			// data[] = %xxxxxxxx'xxxxxxxx'xxxxxxxx'ABCDEFGH
			// ↓
			// (リトルエンディアンホストの場合)
			// p.H = %0000000H'0000000G'0000000F'0000000E
			// p.L = %0000000D'0000000C'0000000B'0000000A
			for (int b = 0; b < 4; b++) {
				uint64 cc = 0;

				for (int plane = 0; plane < nplane4; plane++) {
					uint64 p;
					p = deptable[data[plane] & 0xff];
					p <<= plane;
					cc |= p;
					data[plane] >>= 8;
				}

				// cc (64bit = 8バイト) は各バイトが各ピクセルの
				// カラーインデックスになっているので
				// BitmapI8 に 8バイト一気に書き込める。
				// (エンディアンの影響は deptable[] で処理してある)
				*dst++ = cc;
			}
		}
	}
}

// composite 画面とパレット情報から RGBX 画面を合成する。
// modify は composite 上での変更箇所を示している。
// レンダラスレッドから呼ばれる。
void
PlaneVRAMDevice::RenderCompositeToRGBX(BitmapRGBX& dst,
	const ModifyInfo& modify)
{
	const int width  = dst.GetWidth();
	const int height = dst.GetHeight();

	// view_mod は表示領域 view における更新ラスタ
	BitmapI8 view(width, height);
	std::vector<uint8> view_mod(height);

	// この view に対する modify を計算する
	if (__predict_false(modify.invalidate2)) {
		std::fill(view_mod.begin(), view_mod.end(), 1);
	} else {
		// Y scroll 方向の modify 計算
		int sy = yscroll;
		for (int y = 0; y < height; y++, sy++) {
			if (sy >= composite.GetHeight()) {
				sy -= composite.GetHeight();
			}
			view_mod[y] = modify.bits[sy];
		}

		// 右がはみ出す場合は、はみ出した先のラスタも加味する。
		if (xscroll + view.GetWidth() > composite.GetWidth()) {
			sy = yscroll;
			for (int y = 0; y < height; y++, sy++) {
				if (sy >= composite.GetHeight()) {
					sy -= composite.GetHeight();
				}
				int sy2 = GetWrapY(sy);
				view_mod[y] |= modify.bits[sy2];
			}
		}
	}

	// composite からスクロールを加味した表示領域 view を作る
	int sy = yscroll;
	for (int y = 0; y < height; y++, sy++) {
		if (sy >= composite.GetHeight()) {
			sy -= composite.GetHeight();
		}
		if (view_mod[y] == 0) {
			continue;
		}

		uint8 *v = view.GetRowPtr(y);
		uint8 *c = composite.GetPtr(xscroll, sy);
		if (xscroll + view.GetWidth() <= composite.GetWidth()) {
			memcpy(v, c, view.GetWidth());
		} else {
			// はみ出しの折り返し
			int len1 = composite.GetWidth() - xscroll;
			memcpy(v, c, len1);
			v += len1;
			int sy2 = GetWrapY(sy);
			c = composite.GetRowPtr(sy2);
			int len2 = view.GetWidth() - len1;
			memcpy(v, c, len2);
		}
	}

	// view (I8) を Bitmap (RGBX) に展開する
#if defined(HAVE_AVX2)
	if (__predict_true(enable_avx2)) {
		RenderI8toRGBX_avx2(dst, view, view_mod);
	} else
#endif
	{
		RenderI8toRGBX_gen(dst, view, view_mod);
	}
}

// I8 から RGBX への変換。(C++ 版)
void
PlaneVRAMDevice::RenderI8toRGBX_gen(BitmapRGBX& dst, const BitmapI8& view,
	const std::vector<uint8>& view_mod)
{
	for (int y = 0; y < view_mod.size(); y++) {
		if (view_mod[y] == 0) {
			continue;
		}

		const uint8 *s8 = view.GetRowPtr(y);
		uint32 *d = (uint32 *)dst.GetRowPtr(y);

		const uint32 *s = (const uint32 *)s8;
		const uint32 *send = (const uint32 *)(s8 + view.GetWidth());
		for (; s < send;) {
			// 4ピクセルずつ処理する

			// s は BitmapI8 なのでビッグエンディアン並びに相当。
			uint32 cc4 = *s++;
			// 左側ピクセルが下位バイトになるよう並び替える
			cc4 = htole32(cc4);

			// 4ピクセルをそれぞれ xBGR32 に変換
			uint32 c0 = palette[ cc4        & 0xff].xbgr;
			uint32 c1 = palette[(cc4 >>  8) & 0xff].xbgr;
			uint32 c2 = palette[(cc4 >> 16) & 0xff].xbgr;
			uint32 c3 = palette[ cc4 >> 24        ].xbgr;

			// c0 = $x0B0G0R0
			// c1 = $x1B1G1R1
			// c2 = $x2B2G2R2
			// c3 = $x3B3G3R3
			*d++ = le32toh(c0);
			*d++ = le32toh(c1);
			*d++ = le32toh(c2);
			*d++ = le32toh(c3);
		}
	}
}

==> nono-1.0.0.tar.gz: nono-1.0.0/vm/planevram.cpp <==
//
// nono
// Copyright (C) 2022 nono project
// Licensed under nono-license.txt
//

//
// プレーン VRAM (垂直 VRAM)
// Lunafb と X68k TVRAM の共通部分
//

#include "planevram.h"
#include "mybswap.h"

// コンストラクタ
PlaneVRAMDevice::PlaneVRAMDevice(int width_, int height_)
	: inherited(OBJ_PLANEVRAM)
{
	// ポカ避け。継承クラス側が設定する。
	nplane = -1;

	composite.Create(width_, height_);

	dirty.resize(height_);

	// レンダリング時に使うテーブルを事前に計算。
	InitDeptable();
}

// デストラクタ
PlaneVRAMDevice::~PlaneVRAMDevice()
{
}

// リセット
void
PlaneVRAMDevice::ResetHard(bool poweron)
{
	// XXX ここ?
	Invalidate();
}

// レンダリング時に使うテーブルを事前に計算。
void
PlaneVRAMDevice::InitDeptable()
{
	deptable.reset(new uint64[256]);

	// %abcdefgh の8ビットを
	// リトルエンディアンホストでは
	// %0000000h'0000000g'0000000f'0000000e'0000000d'0000000c'0000000b'0000000a
	// ビッグエンディアンホストでは
	// %0000000a'0000000b'0000000c'0000000d'0000000e'0000000f'0000000g'0000000h
	// の64ビットに伸張する。
	// (8ビットを各バイトの最下位ビットに対応させる)
	for (uint i = 0; i < 256; i++) {
		uint64 res = 0;
		for (int bit = 0; bit < 8; bit++) {
			int n = bit * 8;
#if BYTE_ORDER == LITTLE_ENDIAN
			n = 56 - n;
#endif
			res |= (uint64)((i >> bit) & 1) << n;
		}
		deptable[i] = res;
	}
}

// 全画面の更新フラグを立てる
void
PlaneVRAMDevice::Invalidate()
{
	std::fill(dirty.begin(), dirty.end(), 1);
}

// 全画面の更新フラグを下ろす
// VM スレッドから呼ばれる。
void
PlaneVRAMDevice::ClearDirty()
{
	std::fill(dirty.begin(), dirty.end(), 0);
}

// 更新情報を取得。更新がなければ false を返す。
// VM スレッドから呼ばれる。
bool
PlaneVRAMDevice::GetModify(ModifyInfo *mod) const
{
	memcpy(&mod->bits[0], &dirty[0], mod->bits.size());

	// 一つでも更新があれば dirty。
	mod->dirty = false;
	for (int i = 0; i < dirty.size(); i++) {
		if (dirty[i]) {
			mod->dirty = true;
			break;
		}
	}
	return mod->dirty;
}

// 画面合成。
// レンダラスレッドから呼ばれる。
bool
PlaneVRAMDevice::Render(BitmapRGBX& dst, const ModifyInfo& modify)
{
	bool updated = false;

	// VRAM に更新があれば composite を更新
	if (modify.IsDirty()) {
		RenderVRAMToComposite(modify);
		updated = true;
	}

	// composite に更新があるか(update)、
	// 無条件に更新するか(invalidate2) なら dst を更新。
	if (updated || modify.invalidate2) {
		RenderCompositeToRGBX(dst, modify);
		updated = true;
	}

	return updated;
}

// VRAM から composite 画面を合成する。
// レンダラスレッドから呼ばれる。
void
PlaneVRAMDevice::RenderVRAMToComposite(const ModifyInfo& modify)
{
	const uint width  = composite.GetWidth();
	const uint height = composite.GetHeight();
	const uint planesz = (width / 8) * height;

	// 1bpp はパレット的にはほぼ 4bpp と同じ扱いになる。
	const int nplane4 = (nplane == 1) ? 4 : nplane;

	// テキスト合成画面を描く
	for (uint y = 0; y < height; y++) {
		uint8 mod = modify.bits[y];
		if (mod == 0) {
			continue;
		}

		const uint32 *src = (const uint32 *)&mem[0];
		src += y * width / 32;
		uint64 *dst = (uint64 *)composite.GetRowPtr(y);

		for (uint x = 0; x < width / 32; x++) {
			std::array<uint32, 8> data;

			// VRAM はホストエンディアンに関係なく 32bit で読み込んだ
			// 時の MSB が左端ピクセル。
			// ここでは左側から 8ピクセル(8ビット)ずつ 4回処理する
			// のを最下位バイトから順に右シフトで行いたいので、
			// (バイト内のビット順は維持したまま) バイトスワップする。
			//
			//          31     24  23     16  15      8   7      0
			//        +----------+----------+----------+----------+
			// VRAM   |+00 .. +07|+08 .. +15|+16 .. +23|+24 .. +31|
			//        +----------+----------+----------+----------+
			//  ↓
			//        +----------+----------+----------+----------+
			// data[] |+24 .. +31|+16 .. +23|+08 .. +15|+00 .. +07|
			//        +----------+----------+----------+----------+
			if (__predict_false(nplane == 1)) {
				// LUNA 1bpp の場合残りの3プレーンは ff で埋める。
				// 実際にも存在しないプレーンは ff が読み出せて
				// カラーインデックスは $e か $f の二択になるので。
				data[0] = bswap32(src[0]);
				data[1] = 0xffffffff;
				data[2] = 0xffffffff;
				data[3] = 0xffffffff;
			} else {
				uint offset = planesz / sizeof(uint32);
				for (int plane = 0; plane < nplane; plane++) {
					data[plane] = bswap32(src[plane * offset]);
				}
			}
			src++;

			// data[] の下位 8bit ずつを取り出して deptable[] で変換。
			// p0..p3 は1バイト x 8個。
			//
			// data[] = %xxxxxxxx'xxxxxxxx'xxxxxxxx'ABCDEFGH
			// ↓
			// (リトルエンディアンホストの場合)
			// p.H = %0000000H'0000000G'0000000F'0000000E
			// p.L = %0000000D'0000000C'0000000B'0000000A
			for (int b = 0; b < 4; b++) {
				uint64 cc = 0;

				for (int plane = 0; plane < nplane4; plane++) {
					uint64 p;
					p = deptable[data[plane] & 0xff];
					p <<= plane;
					cc |= p;
					data[plane] >>= 8;
				}

				// cc (64bit = 8バイト) は各バイトが各ピクセルの
				// カラーインデックスになっているので
				// BitmapI8 に 8バイト一気に書き込める。
				// (エンディアンの影響は deptable[] で処理してある)
				*dst++ = cc;
			}
		}
	}
}

// composite 画面とパレット情報から RGBX 画面を合成する。
// modify は composite 上での変更箇所を示している。
// レンダラスレッドから呼ばれる。
void
PlaneVRAMDevice::RenderCompositeToRGBX(BitmapRGBX& dst,
	const ModifyInfo& modify)
{
	const uint width  = dst.GetWidth();
	const uint height = dst.GetHeight();

	// view_mod は表示領域 view における更新ラスタ
	BitmapI8 view(width, height);
	std::vector<uint8> view_mod(height);

	// この view に対する modify を計算する
	if (__predict_false(modify.invalidate2)) {
		std::fill(view_mod.begin(), view_mod.end(), 1);
	} else {
		// Y scroll 方向の modify 計算
		int sy = yscroll;
		for (int y = 0; y < height; y++, sy++) {
			if (sy >= composite.GetHeight()) {
				sy -= composite.GetHeight();
			}
			view_mod[y] = modify.bits[sy];
		}

		// 右がはみ出す場合は、はみ出した先のラスタも加味する。
		if (xscroll + view.GetWidth() > composite.GetWidth()) {
			sy = yscroll;
			for (int y = 0; y < height; y++, sy++) {
				if (sy >= composite.GetHeight()) {
					sy -= composite.GetHeight();
				}
				int sy2 = GetWrapY(sy);
				view_mod[y] |= modify.bits[sy2];
			}
		}
	}

	// composite からスクロールを加味した表示領域 view を作る
	int sy = yscroll;
	for (int y = 0; y < height; y++, sy++) {
		if (sy >= composite.GetHeight()) {
			sy -= composite.GetHeight();
		}
		if (view_mod[y] == 0) {
			continue;
		}

		uint8 *v = view.GetRowPtr(y);
		uint8 *c = composite.GetPtr(xscroll, sy);
		if (xscroll + view.GetWidth() <= composite.GetWidth()) {
			memcpy(v, c, view.GetWidth());
		} else {
			// はみ出しの折り返し
			int len1 = composite.GetWidth() - xscroll;
			memcpy(v, c, len1);
			v += len1;
			int sy2 = GetWrapY(sy);
			c = composite.GetRowPtr(sy2);
			int len2 = view.GetWidth() - len1;
			memcpy(v, c, len2);
		}
	}

	// view (I8) を Bitmap (RGBX) に展開する
	dst.DrawBitmapI8Raster(view, palette, &view_mod[0]);
}
